/*
Copyright 1990-2001 Sun Microsystems, Inc. All Rights Reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions: The above copyright notice and this
permission notice shall be included in all copies or substantial
portions of the Software.


THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE OPEN GROUP OR SUN MICROSYSTEMS, INC. BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE EVEN IF
ADVISED IN ADVANCE OF THE POSSIBILITY OF SUCH DAMAGES.


Except as contained in this notice, the names of The Open Group and/or
Sun Microsystems, Inc. shall not be used in advertising or otherwise to
promote the sale, use or other dealings in this Software without prior
written authorization from The Open Group and/or Sun Microsystems,
Inc., as applicable.


X Window System is a trademark of The Open Group

OSF/1, OSF/Motif and Motif are registered trademarks, and OSF, the OSF
logo, LBX, X Window System, and Xinerama are trademarks of the Open
Group. All other trademarks and registered trademarks mentioned herein
are the property of their respective owners. No right, title or
interest in or to any trademark, service mark, logo or trade name of
Sun Microsystems, Inc. or its licensors is granted.

*/
#ifndef _imtrans_hh
#define _imtrans_hh
#include <string.h>
#ifdef WIN32
#include <windows.h>
#ifdef CYGWIN_NT
#include <sys/time.h>
#endif
#else
#include <sys/time.h>
#endif
#include "CompoundString.hh"
#include <sys/select.h>

class IMTransAccept;

class IMTransListen {
 public:
  IMTransListen(CompoundString im_address, CompoundString protocol_name="",
						int network=1);
#ifdef linux
  /* 4354367: Linix Htt: htt_xbe can not start on standalone system */
  void IMTransListenUNIX(CompoundString im_address, CompoundString protocol_name="");
#endif
  void IMTransListenINET(CompoundString im_address, CompoundString protocol_name="");
  ~IMTransListen();
  IMTransAccept* accept();
  int connection_id() { return listen_socket; }
 private:
  int check_fd_resource_limit(int fd);
  int listen_socket;
  int cur_maxfd;
  int incr_fd;
  int fd_resource_limit;
};

class IMTransAccept {
 public:
  IMTransAccept(int fd) : accept_fd(fd) {
    memset(&rmask_fd, 0, sizeof(fd_set));
    FD_SET(fd, &rmask_fd);
  }
  ~IMTransAccept();
  int read(char *buf, int len);
  int write(char *buf, int len);
  int connection_id();
  int select_fd(struct timeval);

  enum SELECT_TYPE {
    Trans_ERROR = -1, Trans_TIMEOUT = 0, Trans_ACCEPT = 1
  };
 private:
  // dummy declaration to prevent making a copy
  IMTransAccept(const IMTransAccept&);
  IMTransAccept& operator=(const IMTransAccept& a);
  int accept_fd;
  fd_set rmask_fd;
};
#endif
